# Copyright (c) Alibaba, Inc. and its affiliates.
from .ocr_cls_datasource import OCRClsSource
from .ocr_det_datasource import OCRDetSource, OCRPaiDetSource
from .ocr_rec_datasource import OCRReclmdbSource, OCRRecSource
