# Copyright (c) Alibaba, Inc. and its affiliates.
from .cls_head import ClsHead
from .contrastive_head import ContrastiveHead, DebiasedContrastiveHead
from .latent_pred_head import LatentPredictHead
from .mp_metric_head import MpMetrixHead
from .multi_cls_head import MultiClsHead
